@@ -305,6 +305,12 @@ class Daemon(_EntryPoint):
 
         --notify-command COMMAND                Command to execute on events
         --no-notify-command                     Disable command notifications
+
+        --menu-checkbox-workaround              Use checkbox workaround
+        --no-menu-checkbox-workaround           Disable checkbox workaround
+
+        --menu-update-workaround                Use wayland menu workaround
+        --no-menu-update-workaround             Disable wayland menu workaround
     """
 
     option_defaults = extend(_EntryPoint.option_defaults, {
@@ -318,6 +324,8 @@ class Daemon(_EntryPoint):
         'password_prompt': 'builtin:gui',
         'password_cache': False,
         'notify_command': None,
+        'menu_checkbox_workaround': None,
+        'menu_update_workaround': None,
     })
 
     option_rules = extend(_EntryPoint.option_rules, {
@@ -333,6 +341,8 @@ class Daemon(_EntryPoint):
         'password_prompt': OptionalValue('--password-prompt'),
         'password_cache': OptionalValue('--password-cache'),
         'notify_command': OptionalValue('--notify-command'),
+        'menu_checkbox_workaround': OptionalValue('--menu-checkbox-workaround'),
+        'menu_update_workaround': OptionalValue('--menu-update-workaround'),
     })
 
     def _init(self):
@@ -461,6 +471,14 @@ class Daemon(_EntryPoint):
         icons = udiskie.tray.Icons(self.config.icon_names)
         actions = udiskie.mount.DeviceActions(self.mounter)
 
+        checkbox_workaround = options['menu_checkbox_workaround']
+        if checkbox_workaround is None:
+            checkbox_workaround = options['appindicator'] and _in_Wayland
+
+        update_workaround = options['menu_update_workaround']
+        if update_workaround is None:
+            update_workaround = options['appindicator'] and _in_Wayland
+
         if options['menu'] == 'flat':
             flat = True
         # dropped legacy 'nested' mode:
@@ -469,8 +487,11 @@ class Daemon(_EntryPoint):
         else:
             raise ValueError("Invalid menu: %s" % (options['menu'],))
 
-        menu_maker = udiskie.tray.UdiskieMenu(self, icons, actions, flat,
-                                              config.quickmenu_actions)
+        menu_maker = udiskie.tray.TrayMenu(
+            self, icons, actions, flat,
+            config.quickmenu_actions,
+            checkbox_workaround=checkbox_workaround,
+            update_workaround=update_workaround)
         if options['appindicator']:
             import udiskie.appindicator
             TrayIcon = udiskie.appindicator.AppIndicatorIcon
