@@ -25,6 +25,20 @@ def lower(s):
         return s
 
 
+def format_dict(d):
+    return '{' + ', '.join([
+        _format_item(k, v)
+        for k, v in d.items()
+    ]) + '}'
+
+
+def _format_item(k, v):
+    if isinstance(v, bool):
+        return k if v else '!' + k
+    else:
+        return '{}={}'.format(k, v)
+
+
 def match_value(value, pattern):
     if isinstance(value, (list, tuple)):
         return any(match_value(v, pattern) for v in value)
@@ -105,18 +119,19 @@ class DeviceFilter:
             keyfile = os.path.expandvars(keyfile)
             keyfile = os.path.expanduser(keyfile)
             self._values['keyfile'] = keyfile
+        if 'skip' in match:
+            self._values['skip'] = match.pop('skip')
         # the use of list() makes deletion inside the loop safe:
         for k in list(self._match):
             if k not in self.VALID_PARAMETERS:
                 self._log.error(_('Unknown matching attribute: {!r}', k))
                 del self._match[k]
-        self._log.debug(_('{0} created', self))
+        self._log.debug(_('new rule: {0}', self))
 
     def __str__(self):
-        return _('{0}(match={1!r}, value={2!r})',
-                 self.__class__.__name__,
-                 self._match,
-                 self._values)
+        return _('{0} -> {1}',
+                 format_dict(self._match),
+                 format_dict(self._values))
 
     def match(self, device):
         """Check if the device object matches this filter."""
@@ -133,11 +148,8 @@ class DeviceFilter:
         If :meth:`match` is False for the device, the return value of this
         method is undefined.
         """
-        self._log.debug(_('{0}(match={1!r}, {2}={3!r}) used for {4}',
-                          self.__class__.__name__,
-                          self._match,
-                          kind, self._values[kind],
-                          device.object_path))
+        self._log.debug(_('{0} matched {1}',
+                          device.device_file or device.object_path, self))
         return self._values[kind]
 
 
@@ -169,12 +181,17 @@ def match_config(filters, device, kind, default):
     :param default: default value
     :returns: value of the first matching filter
     """
-    if device is None:
-        return default
-    matches = (f.value(kind, device)
-               for f in filters
-               if f.has_value(kind) and f.match(device))
-    return next(matches, default)
+    while device is not None:
+        for f in filters:
+            if f.has_value(kind) and f.match(device):
+                return f.value(kind, device)
+            # 'skip' allows skipping further rules and directly moving on
+            # lookup on the parent device:
+            if f.has_value('skip') and f.match(device) and (
+                    f.value('skip', device) in (True, 'all', kind)):
+                break
+        device = device.partition_slave or device.luks_cleartext_slave
+    return default
 
 
 class Config:
@@ -187,11 +204,10 @@ class Config:
 
     @classmethod
     def default_pathes(cls):
-        """Return the default config file pathes as a list."""
-        try:
-            from xdg.BaseDirectory import xdg_config_home as config_home
-        except ImportError:
-            config_home = os.path.expanduser('~/.config')
+        """Return the default config file paths as a list."""
+        config_home = (
+            os.environ.get('XDG_CONFIG_HOME') or
+            os.path.expanduser('~/.config'))
         return [os.path.join(config_home, 'udiskie', 'config.yml'),
                 os.path.join(config_home, 'udiskie', 'config.json')]
 
