@@ -6,13 +6,16 @@ from gi.repository import Gio
 from gi.repository import Gtk
 
 from .async_ import run_bg, Future
-from .common import setdefault, DaemonBase
+from .common import setdefault, DaemonBase, cachedmethod
 from .locale import _
 from .mount import Action, prune_empty_node
 from .prompt import Dialog
+from .icons import IconDist
+
+import os
 
 
-__all__ = ['UdiskieMenu', 'TrayIcon']
+__all__ = ['TrayMenu', 'TrayIcon']
 
 
 class MenuFolder:
@@ -59,23 +62,32 @@ class Icons:
         'forget_password': ['edit-delete'],
         'delete': ['udiskie-eject'],
         'losetup': ['udiskie-mount'],
+        # checkbox workaround:
+        'checked': ['checkbox-checked', 'udiskie-checkbox-checked'],
+        'unchecked': ['checkbox', 'udiskie-checkbox-unchecked'],
+        'submenu': ['udiskie-submenu', 'pan-end-symbolic'],
     }
 
     def __init__(self, icon_names={}):
         """Merge ``icon_names`` into default icon names."""
+        self._icon_dist = IconDist()
         _icon_names = icon_names.copy()
         setdefault(_icon_names, self.__class__._icon_names)
         self._icon_names = _icon_names
         for k, v in _icon_names.items():
             if isinstance(v, str):
-                self._icon_names[k] = [v]
+                self._icon_names[k] = v = [v]
+            self._icon_names[k] = self._icon_dist.patch_list(v)
 
+    @cachedmethod
     def get_icon_name(self, icon_id: str) -> str:
         """Lookup the system icon name from udisie-internal id."""
         icon_theme = Gtk.IconTheme.get_default()
         for name in self._icon_names[icon_id]:
             if icon_theme.has_icon(name):
                 return name
+            elif os.path.exists(name):
+                return name
         return 'not-available'
 
     def get_icon(self, icon_id: str, size: "Gtk.IconSize") -> "Gtk.Image":
@@ -84,10 +96,17 @@ class Icons:
 
     def get_gicon(self, icon_id: str) -> "Gio.Icon":
         """Lookup Gio.Icon from udiskie-internal id."""
-        return Gio.ThemedIcon.new_from_names(self._icon_names[icon_id])
+        name = self.get_icon_name(icon_id)
+        if os.path.exists(name):
+            # TODO (?): we could also add the icon to the theme using
+            # Gtk.IconTheme.append_search_path or .add_resource_path:
+            file = Gio.File.new_for_path(name)
+            return Gio.FileIcon.new(file)
+        else:
+            return Gio.ThemedIcon.new(name)
 
 
-class UdiskieMenu:
+class TrayMenu:
 
     """
     Builder for udiskie menus.
@@ -96,7 +115,9 @@ class UdiskieMenu:
     """
 
     def __init__(self, daemon, icons, actions, flat=True,
-                 quickmenu_actions=None):
+                 quickmenu_actions=None,
+                 checkbox_workaround=False,
+                 update_workaround=False):
         """
         Initialize a new menu maker.
 
@@ -142,12 +163,29 @@ class UdiskieMenu:
             # 'eject',
             # 'forget_password',
         ]
+        self._checkbox_workaround = checkbox_workaround
+        self._update_workaround = update_workaround
 
     def __call__(self, menu, extended=True):
         """Populate the Gtk.Menu with udiskie mount operations."""
         # create actions items
         flat = self.flat and not extended
-        self._create_menu_items(menu, self._prepare_menu(self.detect(), flat))
+        if self._update_workaround:
+            # When showing menus via AppIndicator3 on sway, the menu geometry
+            # seems to be calculated before the 'about-to-show' event, and
+            # therefore cannot take into account newly inserted menu items.
+            # For this reason, we have to keep the top-level menu fixed-size
+            # and insert dynamic entries into a submenu.
+            devmenu = Gtk.Menu()
+            menu.append(self._menuitem(
+                label=_('Managed devices'),
+                icon=None,
+                onclick=devmenu,
+            ))
+        else:
+            devmenu = menu
+        self._create_menu_items(
+            devmenu, self._prepare_menu(self.detect(), flat))
         if extended:
             self._insert_options(menu)
         return menu
@@ -252,6 +290,13 @@ class UdiskieMenu:
         :returns: the menu item object
         :rtype: Gtk.MenuItem
         """
+        if self._checkbox_workaround:
+            if checked is not None:
+                icon_name = 'checked' if checked else 'unchecked'
+                icon = self._icons.get_icon(icon_name, Gtk.IconSize.MENU)
+                checked = None
+            elif isinstance(onclick, Gtk.Menu):
+                icon = self._icons.get_icon('submenu', Gtk.IconSize.MENU)
         if checked is not None:
             item = Gtk.CheckMenuItem()
             item.set_active(checked)
@@ -319,7 +364,7 @@ class TrayIcon:
         icon will be initially visible, which results in a perceptable
         flickering.
 
-        :param UdiskieMenu menumaker: menu factory
+        :param TrayMenu menumaker: menu factory
         :param Gtk.StatusIcon statusicon: status icon
         """
         self._icons = icons
